/* Sample code using the Chunks and Tasks programming model.
 * See <http://www.chunks-and-tasks.org>.
 */
#include <iostream>
#include <cmath>
#include <cstdlib>
#include "cht_matrix_lib.h"
#include "test_tools.h"


template<typename LeafMatrixType>
int test_lif() {

  // Generate random symmetric positive definite 16x16 matrix with the
  // following nonzero 4x4 block structure:
  //
  // X X X
  // X X
  // X   X X
  //     X X
  //
  unsigned int nnz = 10*4*4;
  chttl::ChunkVector<int>*  rowindex = new chttl::ChunkVector<int>(nnz);
  chttl::ChunkVector<int>*  colindex = new chttl::ChunkVector<int>(nnz);
  chttl::ChunkVector<double>* values = new chttl::ChunkVector<double>(nnz);
  unsigned int counter = 0;
  for (unsigned int row = 0; row < 4; row++) {
    (*rowindex)[counter] = row;
    (*colindex)[counter] = row;
    (*values)[counter]   = 10.0+((double)rand())/RAND_MAX;
    counter++;
    for (unsigned int col = row+1; col < 12; col++) {
      (*rowindex)[counter] = row;
      (*colindex)[counter] = col;
      (*values)[counter]   = ((double)rand())/RAND_MAX;
      counter++;
      (*rowindex)[counter] = col;
      (*colindex)[counter] = row;
      (*values)[counter]   = (*values)[counter-1];
      counter++;
    }
  }
  for (unsigned int row = 4; row < 8; row++) {
    (*rowindex)[counter] = row;
    (*colindex)[counter] = row;
    (*values)[counter]   = 10.0+((double)rand())/RAND_MAX;
    counter++;    
    for (unsigned int col = row+1; col < 8; col++) {
      (*rowindex)[counter] = row;
      (*colindex)[counter] = col;
      (*values)[counter]   = ((double)rand())/RAND_MAX;
      counter++;
      (*rowindex)[counter] = col;
      (*colindex)[counter] = row;
      (*values)[counter]   = (*values)[counter-1];
      counter++;
    }
  }
  for (unsigned int row = 8; row < 16; row++) {
    (*rowindex)[counter] = row;
    (*colindex)[counter] = row;
    (*values)[counter]   = 10.0+((double)rand())/RAND_MAX;
    counter++;    
    for (unsigned int col = row+1; col < 16; col++) {
      (*rowindex)[counter] = row;
      (*colindex)[counter] = col;
      (*values)[counter]   = ((double)rand())/RAND_MAX;
      counter++;
      (*rowindex)[counter] = col;
      (*colindex)[counter] = row;
      (*values)[counter]   = (*values)[counter-1];
      counter++;
    }
  }
  assert(counter == nnz);
  cht::ChunkID cid_rowind = cht::registerChunk<chttl::ChunkVector<int> >(rowindex);
  cht::ChunkID cid_colind = cht::registerChunk<chttl::ChunkVector<int> >(colindex);
  cht::ChunkID cid_values = cht::registerChunk<chttl::ChunkVector<double> >(values);
  
  int M = 16;
  int N = 16;
  int leavesSizeMax = 4;
  int leaf_blocksize = 2;
  typename LeafMatrixType::Params leaf_params;
  set_block_size<LeafMatrixType>(leaf_params, leaf_blocksize);
  cht::ChunkID cid_param = cht::registerChunk<chtml::MatrixParams<LeafMatrixType> >(new chtml::MatrixParams<LeafMatrixType>(M, N, leavesSizeMax, 0, 0, leaf_params));
  cht::ChunkID cid_matrix = 
    cht::executeMotherTask<chtml::MatrixAssignFromSparse<LeafMatrixType> >(cid_param,
									   cid_rowind,
									   cid_colind,
									   cid_values);
  
  cht::ChunkID cid_refinement_m = cht::registerChunk<chttl::Basic<int> >(new chttl::Basic<int>(2));
  cht::ChunkID cid_RIFRefinement_truncation_threshold = cht::registerChunk<chttl::Basic<double> >(new chttl::Basic<double>(1e-6));
  cht::ChunkID cid_rinch_truncation_threshold = cht::registerChunk<chttl::Basic<double> >(new chttl::Basic<double>(0.0));
  cht::ChunkID cid_rinch_matrix_size = cht::registerChunk<chttl::Basic<int> >(new chttl::Basic<int>(1));


  cht::ChunkID MINUS_ONE;
   {
     chttl::Basic<double>* minus_one = new chttl::Basic<double>;
     *minus_one = (double)-1.0;
     MINUS_ONE = cht::registerChunk<chttl::Basic<double> >(minus_one);
   }

  

    cht::ChunkID cid_rif_factor;
    cht::ChunkID cid_inv_matrix;
    cht::ChunkID cid_should_be_identity;
    cht::ChunkID cid_should_be_zero;

    cid_rif_factor = cht::executeMotherTask<chtml::MatrixLocalizedInverseFactorization<LeafMatrixType> >(cid_matrix,
	      cid_param,
	      cid_refinement_m,
	      cid_RIFRefinement_truncation_threshold,
	      cid_rinch_truncation_threshold,
	      cid_rinch_matrix_size);



    cid_inv_matrix =  cht::executeMotherTask<chtml::MatrixMultiply<LeafMatrixType, false, true> >(cid_rif_factor, cid_rif_factor);
    cid_should_be_identity = cht::executeMotherTask<chtml::MatrixMultiply<LeafMatrixType> >(cid_inv_matrix, cid_matrix);
    cid_should_be_zero = cht::executeMotherTask<chtml::MatrixAddScaledIdentity<LeafMatrixType> >(cid_param, cid_should_be_identity, MINUS_ONE);

    double norm_value = chtml::normFrobenius<LeafMatrixType>(cid_should_be_zero);
    std::cout << "Norm is " << norm_value << std::endl;
    assert(norm_value < 1e-4);

    cht::deleteChunk(cid_rif_factor);
    cht::deleteChunk(cid_inv_matrix);
    cht::deleteChunk(cid_should_be_identity);
    cht::deleteChunk(cid_should_be_zero);

  
    cht::deleteChunk(cid_param);
    cht::deleteChunk(cid_rowind);
    cht::deleteChunk(cid_colind);
    cht::deleteChunk(cid_values);  
    cht::deleteChunk(cid_matrix);
    cht::deleteChunk(MINUS_ONE);
    cht::deleteChunk(cid_refinement_m);
    cht::deleteChunk(cid_RIFRefinement_truncation_threshold);
    cht::deleteChunk(cid_rinch_truncation_threshold);
    cht::deleteChunk(cid_rinch_matrix_size);

  return 0;
}

int main() { 
  cht::setOutputLevel(cht::Output::Info);
  cht::start();
  int t1 = test_lif<bml::FullMatrix<double> >();
  if (t1 == 0)
    std::cout << "Test LIF                ( bml::FullMatrix<double> )       : OK!" << std::endl;
  int t2 = test_lif<bsm::BlockSparseMatrix<double> >();
  if (t2 == 0)
    std::cout << "Test LIF                ( bsm::BlockSparseMatrix<double> ): OK!" << std::endl;
  cht::stop(); 
  return t1 || t2;
}
